@@ -1,8 +1,5 @@
 from __future__ import absolute_import, print_function, division, unicode_literals
 
-import six
-from sys import version_info
-
 import pytest
 import requests
 import responses
@@ -18,7 +15,7 @@ def assert_response(resp, body=None, content_type="text/plain"):
 
 
 def assert_reset():
-    assert len(responses._default_mock._matches) == 0
+    assert len(responses._default_mock.registered()) == 0
     assert len(responses.calls) == 0
 
 
@@ -182,11 +179,9 @@ def test_request_matches_params():
         }
         resp = requests.get(url, params=params)
 
-        if six.PY3 and version_info[1] >= 7:
-            # only after py 3.7 dictionaries are ordered, so we can check URL
-            constructed_url = r"http://example.com/test?I+am=a+big+test&hello=world"
-            assert resp.url == constructed_url
-            assert resp.request.url == constructed_url
+        constructed_url = r"http://example.com/test?I+am=a+big+test&hello=world"
+        assert resp.url == constructed_url
+        assert resp.request.url == constructed_url
 
         resp_params = getattr(resp.request, "params")
         assert resp_params == params
@@ -283,17 +278,32 @@ def test_fail_matchers_error():
     assert_reset()
 
 
-def test_multipart_matcher():
+@pytest.mark.parametrize(
+    "req_file,match_file",
+    [
+        (b"Old World!", "Old World!"),
+        ("Old World!", b"Old World!"),
+        (b"Old World!", b"Old World!"),
+        ("Old World!", "Old World!"),
+        (b"\xacHello World!", b"\xacHello World!"),
+    ],
+)
+def test_multipart_matcher(req_file, match_file):
     @responses.activate
     def run():
         req_data = {"some": "other", "data": "fields"}
-        req_files = {"file_name": b"Old World!"}
         responses.add(
             responses.POST,
             url="http://httpbin.org/post",
-            match=[matchers.multipart_matcher(req_files, data=req_data)],
+            match=[
+                matchers.multipart_matcher(
+                    files={"file_name": match_file}, data=req_data
+                )
+            ],
+        )
+        resp = requests.post(
+            "http://httpbin.org/post", data=req_data, files={"file_name": req_file}
         )
-        resp = requests.post("http://httpbin.org/post", data=req_data, files=req_files)
         assert resp.status_code == 200
 
         with pytest.raises(TypeError):
@@ -334,13 +344,14 @@ def test_multipart_matcher_fail():
 
             msg = str(excinfo.value)
             assert "multipart/form-data doesn't match. Request body differs." in msg
+
             assert (
-                '\r\nContent-Disposition: form-data; name="file_name"; '
-                'filename="file_name"\r\n\r\nOld World!\r\n'
+                r'\r\nContent-Disposition: form-data; name="file_name"; '
+                r'filename="file_name"\r\n\r\nOld World!\r\n'
             ) in msg
             assert (
-                '\r\nContent-Disposition: form-data; name="file_name"; '
-                'filename="file_name"\r\n\r\nNew World!\r\n'
+                r'\r\nContent-Disposition: form-data; name="file_name"; '
+                r'filename="file_name"\r\n\r\nNew World!\r\n'
             ) in msg
 
         # x-www-form-urlencoded request
